import click
from mpm.model import *
import pathlib
from mpm.config import ServerParamType

@click.group(help='Add, remove, and synchronize Servers')
def server():
    pass

@server.command(help='Add a new server')
@click.pass_context
@click.argument('name')
@click.argument('path', type=click.Path(dir_okay=True, exists=True))
def add(ctx, name, path):
    ctx.obj['config'].add_server(name, path)
    ctx.obj['config'].save()
    click.echo("Added server {} in {}".format(name, path))

@server.command('list', help='List configured servers')
@click.pass_context
def list_(ctx):
    for server in ctx.obj['config'].servers():
        click.echo('{} ({}):'.format(server.name, server.path))
        serverPath = pathlib.Path(server.path)
        if not serverPath.exists():
            click.echo(click.style("\tServer path does not exist", fg='red'))
            continue
        pluginPath = pathlib.Path(server.path+'/plugins')
        if not pluginPath.exists():
            click.echo(click.style("\tServer plugin directory does not exist", fg='red'))
            continue
        outdatedLinks = []
        missing = []
        installed = []
        unmanaged = []
        conflicts = []
        for state in sorted(server.pluginStates(ctx.obj['config'].repositories())):
            if isinstance(state, OutdatedSymlink):
                outdatedLinks.append(state)
            elif isinstance(state, Installed):
                installed.append(state)
            elif isinstance(state, MissingVersions):
                missing.append(state)
            elif isinstance(state, UnmanagedFile):
                unmanaged.append(state)
            elif isinstance(state, SymlinkConflict):
                conflicts.append(state)

        click.echo("Installed plugins:")
        for state in sorted(installed):
            click.echo("\t{} {}: {}".format(state.plugin.name, state.plugin.versionSpec, state.currentVersion))
        click.echo("Oudated symlinks:")
        for state in sorted(outdatedLinks):
            click.echo("\t{} {}: Current: {} Wanted: {}".format(state.plugin.name, state.plugin.versionSpec, state.currentVersion, state.wantedVersion))
        click.echo("Missing plugins:")
        for state in sorted(missing):
            click.echo("\t{}: {}".format(state.plugin.name, state.plugin.versionSpec))
        click.echo("Unmanaged files:")
        for state in sorted(unmanaged):
            click.echo("\t{}".format(state.filename))
        click.echo("Symlink Conflicts:")
        for state in sorted(conflicts):
            click.echo("\t{}.jar".format(state.plugin.name))

@server.command(help='Add a plugin to a server')
@click.pass_context
@click.argument('server', type=ServerParamType())
@click.argument('plugins', nargs=-1)
def add_plugin(ctx, server, plugins):
    pluginQueue = []
    for pluginSpec in plugins:
        if os.path.exists(pluginSpec):
            plugin = Plugin(pluginSpec)
            pluginSpec = PluginSpec(plugin.name, str(plugin.version))
        else:
            allVersions = []
            for repo in ctx.obj['config'].repositories():
                allVersions += repo.versionsForPlugin(pluginSpec)
            pluginSpec = PluginSpec(pluginSpec, list(reversed(sorted(allVersions)))[0])

        pluginQueue.append(pluginSpec)

    click.echo("Resolved plugin list:")
    for pluginSpec in pluginQueue:
        click.echo("\t{} {}".format(click.style(pluginSpec.name, fg='green'),
                                    click.style(pluginSpec.versionSpec,
                                    fg='yellow')))
    click.echo("Add these plugins to server {}? [y/N] ".format(server.name),
               nl=False)
    answer = click.getchar().lower()
    if answer == "y":
        for pluginSpec in pluginQueue:
            try:
                server.add_plugin(pluginSpec)
            except KeyError:
                click.echo("{} already added!".format(pluginSpec))
        ctx.obj['config'].update_server(server.name, server.config)
        ctx.obj['config'].save()
        click.echo("Added!")
    else:
        click.echo("Cancelled.")


@server.command(help="Synchronize a server's plugins")
@click.pass_context
@click.argument('servers', type=ServerParamType, nargs=-1)
def sync(ctx, servers):
    allServers = ctx.obj['config'].servers()
    if len(servers) > 0:
        allServers = servers
    for server in allServers:
        click.echo('{} ({}):'.format(server.name, server.path))
        outdatedLinks = []
        available = []
        missing = []
        for state in sorted(server.pluginStates(ctx.obj['config'].repositories())):
            if isinstance(state, OutdatedSymlink):
                outdatedLinks.append(state)
            elif isinstance(state, Available):
                available.append(state)
            elif isinstance(state, MissingVersions):
                missing.append(state)

        click.echo("Plugins to update:")
        for state in sorted(outdatedLinks):
            click.echo("\t{} {}: Current: {} Wanted: {}".format(state.plugin.name, state.plugin.versionSpec, state.currentVersion, state.wantedVersion))
        click.echo("New plugins to install:")
        for state in sorted(available):
            click.echo("\t{}: {}".format(state.plugin.name, state.plugin.version))
        click.echo("Missing plugins:")
        for state in sorted(missing):
            click.echo("\t{} {}".format(click.style(state.plugin.name, fg='red'),
                                   click.style(str(state.plugin.versionSpec),
                                   fg='yellow')))
        if len(outdatedLinks) > 0 or len(available) > 0:
            click.echo("Apply changes? [y/N]", nl=False)
            answer = click.getchar().lower()
            if answer == "y":
                with click.progressbar(length=len(available)+len(outdatedLinks),
                                       label='Synchronizing') as bar:
                    i = 0
                    for state in available:
                        server.installVersion(state.plugin)
                        server.updateSymlinkForPlugin(state.plugin, state.plugin.version)
                        i += 1
                        bar.update(i)
                    for state in outdatedLinks:
                        server.updateSymlinkForPlugin(state.plugin, state.wantedVersion)
                        i += 1
                        bar.update(i)
            else:
                click.echo("Not applying changes.")
        else:
            click.echo("No changes to apply.")
